; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Sine/Cosine                                *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FSIN

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   RFrac        : FAR
             EXTRN   RealAdd      : NEAR
             EXTRN   RealFloat    : NEAR
             EXTRN   RealMul      : NEAR
             EXTRN   RealMulFNoChk: NEAR
             EXTRN   RealMulNoChk : NEAR
             EXTRN   RealPoly     : NEAR
             EXTRN   RealReduceMP : NEAR
             EXTRN   RealSub      : NEAR
             EXTRN   RealTrunc    : NEAR

; Publics

             PUBLIC  RSin
             PUBLIC  RCos

;-------------------------------------------------------------------------------
; RSin and RCos are different entries into the same routine that computes the
; sine as well as the cosine, as cosine is the same as the sine with an offset
; of pi/2. The argument is reduced to the range -pi/4..+pi/4. Depending on the
; octant into which the argument falls and the function to be computed, the
; cosine or sine of the reduced argument is computed by polynomial approxi-
; mations to the sine/cosine. To prevent excessive loss of precision in the
; argument reduction, the mapping of the argument to the reduced argument is
; done via a pseudo extended precision calculation. However this extended
; precision calculation is only available for arguments inside the range
; -1.68e+9..1.68e+9. Outside this range, a simple argument reduction is used.
; This causes a sudden drop in accuracy if the absolute value of the argument
; gets larger than 1.68e+9. For arguments bigger than 6.9e12, there is a total
; loss of precision. Outside this range, Sin always returns 0, Cos returns 1.
; The approximating polynomial for the sine on -pi/4..pi/4 is:
;
; P(x) := ((((2.476053850842e-8 *x + 2.755533965768e-6)*x -
;             1.984126372865e-4)*x + 8.333333325083e-3)*x -
;             1.666666666663e-1)*x*x + x
;
; This approximation has a theoretical maximum relative error of 1.13e-14.
; Maximum observed error when evaluated in REAL arithmetic is 1.04e-13.
;
; The approximating polynomial for the cosine on -pi/4..pi/4 is:
;
; P(x) := ((((-2.715314622037e-7 *x + 2.479862035332e-5)*x -
;              1.388887879411e-3)*x + 4.166666651281e-2)*x -
;              4.999999999923e-1)*x + 1.000000000000e+0
;
; This approximation has a theoretical maximum relative error of 1.41e-13.
; Maximum observed error when evaluated in REAL arithmetic is 1.49e-12.
;
; INPUT:     DX:BX:AX  argument
;
; OUTPUT:    DX:BX:AX  sin of argument
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RCos         PROC    FAR

             push    bp                ; save caller's frame pointer
             and     dh, 7fh           ; abs(x), since cos (x) = cos (abs(x))
             mov     bp, 2             ; load function code for cosine
             jmp     $cos_entry        ; goto sine/cosine computation

RCos         ENDP

             INCLUDE ALIGN.ASM

RSin         PROC    FAR

             push    bp                ; save caller's frame pointer
             xor     bp, bp            ; load function code for sine

$cos_entry:  mov     ch, 80h           ; load mask to extract sign bit
             and     ch, dh            ; extract sign bit
             xor     dh, ch            ; x := abs(x)
             push    cx                ; save sign

$value_ok:   mov     cx, 04e81h        ; load
             mov     si, 0836eh        ;  real constant
             mov     di, 022f9h        ;   4/pi
             cmp     al, 9fh           ; x >= 2^30 ?
             jae     $big_val          ; x/(pi/4) too big to be stored in long

             push    ax                ; save x
             push    bx                ;  on
             push    dx                ;   stack
             call    RealMulFNoChk     ; compute x/(pi/4)

             xor     ch, ch            ; signal truncation desired
             call    RealTrunc         ; n = Trunc (x/(pi/4))

             ror     al, 1             ; if quadrant odd, set CF=1
             rol     al, 1             ; restore register, CF=1 if odd quadrant
             adc     ax, 0             ; increment quadrant
             adc     dx, 0             ;  if quadrant odd
             add     bp, ax            ; q := quadrant + flag
             call    RealFloat         ; convert quadrant to REAL value

             pop     di                ; get
             pop     si                ;  back
             pop     cx                ;   x
             test    bp, 3             ; determine octants that use cos,sin
             pushf                     ; save sin/cos flag
             push    bp                ; save q
             mov     bp, offset C1     ; pointer to reduction constant table
             call    RealReduceMP      ; compute x + n*(pi/2) to full precision

             mov     cx, 5             ; use five coefficients
             xor     si, si            ; signal P(x)*x+x wanted
             pop     bp                ; get back q
             popf                      ; get back sin/cos flag
             jpe     $sine             ; octants 0, 3, 4, 7 take sine

$cosine:     mov     di, offset CS_COEFF; pointer to 1st coeff. of cos approx.
             inc     si                ; signal P(x) wanted
             call    RealPoly          ; P(x) in DX:BX:AX

             mov     cx, 00081h        ; load
             xor     si, si            ;  floating-point
             mov     di, si            ;   constant 1.0
             call    RealAdd           ; 1 + P(x)

             jmp     $make_sign        ; put appropriate sign on result

$sine:       mov     di, offset SN_COEFF; pointer to 1st coeff. of sin approx.
             call    RealPoly          ; P(x)*x+x in DX:BX:AX

$make_sign:  pop     cx                ; get sign mask
             inc     bp                ; determine if
             test    bp, 4             ;  result has to negated
             jz      $ende             ; result ok

             xor     ch, 80h           ; negate sign flag

$ende:       pop     bp                ; restore caller's frame pointer
             test    al, al            ; result zero ?
             jz      $end_sin          ; yes, done

             xor     dh, ch            ; make sign bit

$end_sin:    ret                       ; done

$big_val:    sub     cl, 3             ; 1/(2*pi) in DI:SI:CX
             call    RealMulNoChk      ; x/(2*pi)

             call    RFrac             ; frac (x/(2*pi))

             mov     cx, 02183h        ; load
             mov     si, 0daa2h        ;  constant
             mov     di, 0490fh        ;   2*pi
             call    RealMulNoChk      ; 2*pi * frac (x/(2*pi)) = remainder

             jmp     $value_ok         ; reduction to 0..2*pi done

RSin         ENDP

SN_COEFF     db      067h,               0b1h,0d4h  ; -2.476053850842e-8
             db      06eh,05ch,08bh,0b6h,0ebh,038h  ;  2.755533965768e-6
             db      074h,0b5h,09ch,0fch,00ch,0d0h  ; -1.984126372865e-4
             db      07ah,044h,086h,088h,088h,008h  ;  8.333333325083e-3
             db      07eh,0a9h,0aah,0aah,0aah,0aah  ; -1.666666666663e-1

CS_COEFF     db      06bh,               0c7h,091h  ; -2.715314622037e-7
             db      071h,034h,0b7h,0a1h,006h,050h  ;  2.479862035332e-5
             db      077h,02eh,00ah,058h,00bh,0b6h  ; -1.388887879411e-3
             db      07ch,018h,0a0h,0aah,0aah,02ah  ;  4.166666651281e-2
             db      07fh,0efh,0ffh,0ffh,0ffh,0ffh  ; -4.999999999923e-1

C1           db      080h,000h,000h,000h,00fh,0c9h  ; pi/4-eps =
C2           db      070h,0c2h,068h,021h,0a2h,0dah  ;      eps =

             INCLUDE ALIGN.ASM

CODE         ENDS

             END